import std.string;
import std.regexp : RegExp;
import std.conv : to;

//////////////////////////////////////////////////////////////////////////////////////
// 字句解析したトークンを適当に保持する
//////////////////////////////////////////////////////////////////////////////////////

class Token
{
	int    line;
	string value;

	this(int l, string v) { line=l; value=v; }
	string toString() { return "[" ~ to!(string)(line) ~ ": " ~ value ~ "]"; }
}

//////////////////////////////////////////////////////////////////////////////////////
// 字句解析する。文字列渡すとトークンの列にして返す
//////////////////////////////////////////////////////////////////////////////////////

Token[] lex( string src )
{
	// そこまでで強制的にトークンを切りたいものはこの正規表現に突っ込む
	// これに入ってないやつは空白まで全部ひとつながりとトークンということにする
	auto delims = RegExp( `>>|--|->|;;|<=|>=|==|!=|[^_A-Za-z0-9]` );
	bool wassomething = false;

	Token[] result;
	foreach(lineno, s; src.split("\n"))
	{
		s = s.expandtabs(4);
		int indent = s.munch(" ").length; // 一応インデント記録
		bool zeroi = (indent==0 && wassomething);

		while( (s=s.stripl()).length )
		{
			if(int v = delims.find(s))
			{
				// 手抜きオフサイドルールもどき
				if(zeroi) { result ~= new Token(lineno+1, ";;"); zeroi = false; }
				wassomething = true;

				if( v == -1 ) v = s.length;
				result ~= new Token(lineno+1, s[0..v]);
				s = s[v..$];
			}
			else
			{
				if( delims[0] == "--" ) // 行コメ
					break;

				// 手抜きオフサイドルールもどき
				if(zeroi) { result ~= new Token(lineno+1, ";;"); zeroi = false; }
				wassomething = true;

				result ~= new Token(lineno+1, delims[0]);
				s = delims.post;
			}
		}
	}

	result ~= new Token(-1, ";;");
	return result;
}
