--------------------------------------------------------------
-- できること
--  * 型はない
--  * 遅延評価
--  * パターンマッチ
--  * 64bit整数演算
--  * print と >> だけできる IO
--
-- 細かい構文的な色々
--  * [タグ名 値...] でタグ付きの値を作れます
--  * data _ = A | B _ | C _ _ みたいなのは
--      A = [A]
--      B = \x -> [B x]
--      C = \x y -> [C x y]
--    の略記法です。型ないので型っぽいとこは無視
--  * ブールは [True] と [False] を使います
--    組み込みの比較演算子とかもこれを返します
--  * IO もタグで実現されてます
--    [PRINT x] が x を表示するアクション
--    [SEQIO x y] が x と y を順番に実行するアクション
--  * パターンマッチはタグ付き値でのみできます。
--    整数に対する定数パターンは使えません。
--  * ２項演算子は全部左結合です。1+2*3 は 9 です。
---------------------------------------------------------------

-- 組み込みIO
print x = [PRINT x]
nop = [NOP]

-- ブール型とif
data Bool = True | False
if True  t e = t
if False t e = e

-- リスト
data List = Cons _ _ | Nil

car (Cons a d) = a
cdr (Cons a d) = d
caddr (Cons _ (Cons _ (Cons a d))) = a -- 深いパターンマッチのてすと

map fn (Nil)      = Nil
map fn (Cons a d) = Cons (fn a) (map fn d)

take n lst = if (n==0) Nil (Cons (car lst) (take (n-1) (cdr lst)))

fold nl cs Nil        = nl
fold nl cs (Cons a b) = cs a (fold nl cs b)

printlist = fold nop (\a -> \b -> print a >> b)

-- 再帰のテスト
fact x = if (x<1) 1 (x * fact (x-1))

-- 無限リストのテスト
nats = Cons 0 (map (\x -> x+1) nats) 

-- たらい回し
tarai x y z = if (x <= y) y (tarai (tarai (x-1) y z) (tarai (y-1) z x) (tarai (z-1) x y))

-- FizzBuzz
fizzbuzz n = if (n%15 == 0) (print [FizzBuzz])
                            (if (n%3 == 0) (print [Fizz])
                                           (if (n%5 == 0) (print [Buzz]) (print n))) >> fizzbuzz (n+1)

-- ふぃぼなっち
zip f (Cons ah at) (Cons bh bt) = Cons (f ah bh) (zip f at bt)
fibs = Cons 1 (Cons 1 (zip (\x -> \y -> x+y) fibs (cdr fibs)))


fibg = \x0->\x1->case [Tuple x0 x1] of {(Tuple f n) -> if (n<2) 1 (f (n-1) + f (n-2))}
y f = (\x -> f (x x)) (\x -> f (x x))
fib = y fibg

-- てすとてすと
main = print (fib 5)
main = printlist (take 10 fibs)
    >> print (tarai 12 6 0)
    >> print (fact 5 + caddr (cdr (cdr (cdr (cdr nats)))))
    >> print 456
    >> fizzbuzz 0
